"""Verifies the 'snomedct.py' module"""

from urllib.parse import urlparse

from nose.tools import ok_, raises

from petl import Record

from termlink.snomedct import Service, _to_json


def test_service_uri_can_be_file():
    """Checks that a uri.scheme of 'file' is ok"""
    uri = urlparse("file://")
    ok_(Service(uri=uri))


@raises(ValueError)
def test_service_uri_requires_scheme_file():
    """Checks that a uri.scheme of 'foobar' throws a ValueError"""
    uri = urlparse("foobar://")
    Service(uri=uri)


def test_to_json():
    """Checks that a record is properly converted to .json"""

    fields = ['effectiveTime', 'sourceId', 'destinationId', 'source.term', 'target.term', 'versioned']
    values = ('20160131', '425630003', '400195000', 'Acute irritant contact dermatitis (disorder)',
              'Contact hypersensitivity reaction (disorder)', False)
    record = Record(values, fields)
    actual = _to_json(record)

    ok_(actual[0])



"""Handles Snomed-CT conversion.

This module provides methods to extract, transform and load relationships
defined by the SNOMED CT dataset.

The download files for SNOMED CT are provided at https://www.nlm.nih.gov/healthit/snomedct/.
"""
import petl as etl
import json
import os

from termlink.commands import SubCommand
from termlink.models import Coding, Relationship, RelationshipSchema
from termlink.services import RelationshipService
from os import listdir
from dataclasses import replace
from urllib.parse import urlparse

_STATED_RELATIONSHIP_FILE_FIELDS = ["id", "effectiveTime", "active", "moduleId", "sourceId", "destinationId",
                                    "relationshipGroup", "typeId", "characteristicTypeId", "modifierId"]
_DESCRIPTION_FILE_FIELDS = ["id", "effectiveTime", "active", "moduleId", "conceptId", "languageCode", "typeId",
                            "term", "caseSignificanceId"]
_IS_A_TYPE_ID = "116680003"
_FULLY_SPECIFIED_NAME_TYPE_ID = "900000000000003001"
_FIRST_LETTER_CAPITAL_CASE_ID = "900000000000020002"


def _to_json(rec):
    """
    Convert record in table to Relationship as a JSON object

    Record is expected to have the following fields: [ source.CODE, source.STR, target.CODE, target.STR]

    Args:
        rec: A table record

    Returns:
        A new record containing a single field, which is the JSON object
    """

    source = Coding(
        system="http://snomed.info/sct",
        code=rec['sourceId'],
        display=rec['source.term'],
        version=rec['effectiveTime'] if rec['versioned'] else None
    )

    target = Coding(
        system="http://snomed.info/sct",
        code=rec['destinationId'],
        display=rec['target.term'],
        version=rec['effectiveTime'] if rec['versioned'] else None
    )

    relationship = Relationship('subsumes', source, target)

    schema = RelationshipSchema()
    return [json.dumps(schema.dump(relationship))]


class Command(SubCommand):
    """
    A command executor for Snomed-CT operations
    """

    @staticmethod
    def execute(args):
        """
        Prints a JSON array of `Relationship` objects to stdout

        Args:
            args: `argparse` parsed arguments
        """
        uri = urlparse(args.uri)
        service = Service(uri, args.versioned, args.include_inactive)
        table = service.get_relationships()
        if table:
            etl.io.totext(table, encoding='utf8', template='{relationship}\n')



class Service(RelationshipService):
    """Converts the SNOMED CT database"""

    def __init__(self, uri, versioned=False, include_inactive=False):
        """
        Bootstraps a service

        Args:
            uri: URI to root location of .rrf files
            versioned: Flag to indicate whether to include version mapping
            include_inactive: Flag to indicate whether to transform inactive relationship or not. Default is active only.
        """

        if uri.scheme != 'file':
            raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

        self.uri = uri
        self.versioned = versioned
        self.include_inactive = include_inactive

    def get_file_name(self, files, file_type):
        for file in files:
            if file_type in file:
                return os.path.join(self.uri.path, file)
        raise ValueError("Unable to find \"" + file_type + "\" file in the directory files (" + str(files) + ")!")

    def get_relationships(self):
        """
        Parses a list of `Relationship` objects.
        """
        files = listdir(self.uri.path)

        path = self.get_file_name(files, "StatedRelationship")
        relationship = etl \
            .fromcsv(path, delimiter='\t') \
            .setheader(_STATED_RELATIONSHIP_FILE_FIELDS) \
            .select(lambda rec: rec['typeId'] == _IS_A_TYPE_ID)

        # Include inactive if requested
        if not self.include_inactive:
            relationship = relationship \
                .select(lambda rec: rec['active'] == '1')

        relationship = relationship \
            .groupselectlast(['sourceId', 'destinationId', 'typeId']) \
            .cut('effectiveTime', 'sourceId', 'destinationId')

        if relationship.nrows() == 0:
            return None

        path = self.get_file_name(files, "Description")
        description = etl \
            .fromcsv(path, delimiter='\t') \
            .setheader(_DESCRIPTION_FILE_FIELDS) \
            .select(lambda rec: rec['typeId'] == _FULLY_SPECIFIED_NAME_TYPE_ID) \
            .select(lambda rec: rec['caseSignificanceId'] == _FIRST_LETTER_CAPITAL_CASE_ID) \
            .sort(['conceptId', 'effectiveTime'], reverse=True) \
            .cut('conceptId', 'term')

        relationship_with_description = relationship \
            .lookupjoin(description, lkey="sourceId", rkey="conceptId", rprefix="source.") \
            .lookupjoin(description, lkey="destinationId", rkey="conceptId", rprefix="target.") \
            .addfield('versioned', self.versioned) \
            .rowmap(_to_json, ['effectiveTime', 'sourceId', 'destinationId', 'source.term', 'target.term',
                               'versioned']) \
            .setheader(['relationship'])

        return relationship_with_description



"""Handles RxNorm conversion.

This module provides methods to extract, transform and load relationships
defined by the RxNorm dataset.

The download files for RxNorm are provided at https://www.nlm.nih.gov/research/umls/rxnorm/.
"""
import os
import json

from urllib.parse import urlparse

import petl as etl

from termlink.commands import SubCommand
from termlink.models import Coding, Relationship, RelationshipSchema
from termlink.services import RelationshipService

_RXNCONSO_FIELDS = ["RXCUI", "LAT", "TS", "LUI", "STT", "SUI", "ISPREF", "RXAUI",
                    "SAUI", "SCUI", "SDUI", "SAB", "TTY", "CODE", "STR", "SRL", "SUPPRESS", "CVF", ]
_RXNREL_FIELDS = ["RXCUI1", "RXAUI1", "STYPE1", "REL", "RXCUI2", "RXAUI2",
                  "STYPE2", "RELA", "RUI", "SRUI", "SAB", "SL", "DIR", "RG", "SUPPRESS", "CVF", ]

_RELATIONSHIP_TO_EQUIVALENCE = {
    'RB': 'subsumes',
}

_SAB_TO_SYSTEM = {
    'ATC': 'http://www.whocc.no/atc',
    'CVX': 'http://hl7.org/fhir/sid/cvx',
    'SNOMEDCT_US': 'http://snomed.info/sct',
}


def _to_equivalence(rel):
    """Converts a relationship into an equivalence

    Args:
        rel: relationship of second concept or atom to first concept or atom

    Returns:
        an equivalence from https://www.hl7.org/fhir/valueset-concept-map-equivalence.html
    """
    try:
        return _RELATIONSHIP_TO_EQUIVALENCE[rel]
    except KeyError:
        raise RuntimeError('rel \'%s\' is not supported' % rel)


def _to_system(sab):
    """Converts an SAB to a system.

    Args:
        sab: a UMLS source name abbreviation

    Returns:
        the identity of the terminology system
    """
    return _SAB_TO_SYSTEM.get(sab, 'http://www.nlm.nih.gov/research/umls/%s' % sab.lower())


def _to_relationship(rec):
    """Convert record in table to `Relationship`

    Record is expected to have the following fields: [source.CODE, source.STR, source.SAB, target.CODE, target.STR, target.SAB]

    Args:
        rec: A table record

    Returns:
        A `Relationship`
    """
    equivalence = _to_equivalence(rec['REL'])

    source = Coding(
        system=_to_system(rec['source.SAB']),
        code=rec['source.CODE'],
        display=rec['source.STR']
    )

    target = Coding(
        system=_to_system(rec['target.SAB']),
        code=rec['target.CODE'],
        display=rec['target.STR']
    )

    return Relationship(equivalence, source, target)


def _to_json(rec):
    """Converts a record to a formatted `Relationship` in JSON form.

    Args:
        rec: A table record

    Returns:
        A new record containing a single field, which is the JSON object
    """
    relationship = _to_relationship(rec)
    schema = RelationshipSchema()
    return [json.dumps(schema.dump(relationship))]


class Command(SubCommand):
    "A command executor for RxNorm operations"

    @staticmethod
    def execute(args):
        """Prints a JSON array of `Relationship` objects to stdout

        Args:
            args: `argparse` parsed arguments
        """
        uri = urlparse(args.uri)
        vocabularies = set(args.vocabulary)
        suppress_flags = set(args.suppress)
        service = Service(uri, vocabularies, suppress_flags)
        table = service.get_relationships()
        etl.io.totext(table, encoding='utf8', template='{relationship}\n')


class Service:
    """Converts the RxNorm database"""

    def __init__(self, uri, vocabularies=set(['RXNORM']), suppress_flags=set(['N'])):
        """Bootstraps a service

        Args:
            uri: URI to root location of .rrf files
            vocabularies: source vocabularies
            suppress_flags: suppress flags
        """

        if uri.scheme != 'file':
            raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

        self.uri = uri
        self.vocabularies = vocabularies
        self.suppress_flags = suppress_flags

    def get_relationships(self):
        "Parses a list of `Relationship` objects."
        path = os.path.join(self.uri.path, 'RXNCONSO.RRF')
        rxnconso = etl \
            .fromcsv(path, delimiter='|') \
            .setheader(_RXNCONSO_FIELDS) \
            .select(lambda rec: rec['SAB'] in self.vocabularies) \
            .select(lambda rec: rec['SUPPRESS'] in self.suppress_flags) \
            .cut('RXCUI', 'CODE', 'STR', 'SAB')

        source = rxnconso.prefixheader('source.')
        target = rxnconso.prefixheader('target.')

        path = os.path.join(self.uri.path, 'RXNREL.RRF')
        rxnrel = etl \
            .fromcsv(path, delimiter='|') \
            .setheader(_RXNREL_FIELDS) \
            .select(lambda rec: rec['SAB'] in self.vocabularies) \
            .select(lambda rec: rec['STYPE1'] == 'CUI') \
            .select(lambda rec: rec['STYPE2'] == 'CUI') \
            .select(lambda rec: rec['REL'] in _RELATIONSHIP_TO_EQUIVALENCE.keys()) \
            .cut('REL', 'RXCUI1', 'RXCUI2')

        return rxnrel \
            .join(source, lkey='RXCUI1', rkey='source.RXCUI') \
            .join(target, lkey='RXCUI2', rkey='target.RXCUI') \
            .rowmap(_to_json, ['REL', 'source.CODE', 'source.STR', 'source.SAB', 'target.CODE', 'target.STR', 'target.SAB']) \
            .setheader(['relationship'])



"""The main program.

This module is the command line interface for running "termlink."
"""
import argparse
import textwrap

from termlink import common
from termlink import gsea
from termlink import codesystem

from termlink.configuration import Config

from termlink.rxnorm import Command as RxNormCommand
from termlink.hpo import Command as HPOCommand
from termlink.snomedct import Command as SnomedCtCommand
from termlink.loinc import Command as LoincCommand

configuration = Config()
logger = configuration.logger

parser = argparse.ArgumentParser(
    description="""
    An ontology conversion toolkit for the Precision Health Cloud.
    """
)

subparsers = parser.add_subparsers(
    title="Commands",
    metavar=""
)

parser_common = subparsers.add_parser(
    "common",
    help="Convert a common format ontology",
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=textwrap.dedent("""
    Convert an ontology represented in a common format.

    Supported format types:
      - OBO (obo)
      - RDF/XML (owl)
    """)
)

parser_common.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_common.add_argument(
    "-s",
    "--system",
    help="identity of the terminology system",
    required=True
)

parser_common.set_defaults(execute=common.execute)

parser_hpo = subparsers.add_parser(
    "hpo",
    help="Convert the 'Human Phenotype Ontology'",
    description="""
    The Human Phenotype Ontology (HPO) project provides an ontology of medically relevant phenotypes, disease-phenotype annotations, and the algorithms that operate on these. The HPO can be used to support differential diagnostics, translational research, and a number of applications in computational biology by providing the means to compute over the clinical phenotype. The HPO is being used for computational deep phenotyping and precision medicine as well as integration of clinical data into translational research. Deep phenotyping can be defined as the precise and comprehensive analysis of phenotypic abnormalities in which the individual components of the phenotype are observed and described. The HPO is being increasingly adopted as a standard for phenotypic abnormalities by diverse groups such as international rare disease organizations, registries, clinical labs, biomedical resources, and clinical software tools and will thereby contribute toward nascent efforts at global data exchange for identifying disease etiologies (Köhler et al, 2017). [1]
    """,
    epilog="""
    [1] Human Phenotype Ontology. Retrieved April 29, 2019, from https://hpo.jax.org/app/help/introduction
    """
)

parser_hpo.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_hpo.add_argument(
    "--skip-alt-ids",
    dest='skip_alt_ids',
    action='store_true',
    help="skips 'alt_id' references"
)

parser_hpo.add_argument(
    "--skip-synonyms",
    dest='skip_synonyms',
    action='store_true',
    help="skips 'synonym' references"
)

parser_hpo.set_defaults(execute=HPOCommand.execute)
parser_hpo.set_defaults(skip_alt_ids=False)
parser_hpo.set_defaults(skip_synonyms=False)


parser_rxnorm = subparsers.add_parser(
    "rxnorm",
    help="Convert the 'RxNorm' code system",
    description="""
    RxNorm provides normalized names for clinical drugs and links its names to
    many of the drug vocabularies commonly used in pharmacy management and drug
    interaction software, including those of First Databank, Micromedex, 
    Gold Standard Drug Database, and Multum. By providing links between these 
    vocabularies, RxNorm can mediate messages between systems not using the 
    same software and vocabulary. [1] 

    Additional source vocabularies, aside from RxNorm, must be specified using 
    the --vocabulary option. The full list of available source vocabularies is 
    available at https://www.nlm.nih.gov/research/umls/rxnorm/docs/2019/rxnorm_doco_full_2019-1.html#s3_0.

    By default concepts that are not suppressed, specified by the 'N' flag, are
    included in the output. Additional flags can be included using the --suppress
    option. More information on suppression can is documented at https://www.nlm.nih.gov/research/umls/rxnorm/docs/2019/rxnorm_doco_full_2019-1.html#s12_0.
    """,
    epilog="""
    [1] RxNorm. Retrieved April 22, 2019, from https://www.nlm.nih.gov/research/umls/rxnorm/
    """
)

parser_rxnorm.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_rxnorm.add_argument(
    "-v",
    "--vocabulary",
    action='append',
    default=['RXNORM'],
    help="use an additional source vocabulary (default: 'RXNORM')"
)

parser_rxnorm.add_argument(
    "-s",
    "--suppress",
    action='append',
    default=['N'],
    help="unset the provided suppress flag (default: 'N')"
)

parser_rxnorm.set_defaults(execute=RxNormCommand.execute)

parser_gsea = subparsers.add_parser(
    "gsea",
    help="Convert the 'Gene Set Enrichment Analysis Ontology'",
    description="""
    The Gene Set Enrichment Analysis Ontology (GSEA) project provides an ontology of genes grouped by a 
    relational concept. [1]
    """,
    epilog="""
    [1] Gene Set Enrichment Analysis Ontology. Retrieved May 8, 2019, from http://software.broadinstitute.org/gsea/msigdb/collections.jsp
    """
)

parser_gsea.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_gsea.add_argument(
    "--output",
    dest='output',
    help="name of file to generate",
    required=False
)

parser_gsea.set_defaults(execute=gsea.execute)

parser_snomedct = subparsers.add_parser(
    "snomed-ct",
    help="Convert the 'SNOMED-CT' code system",
    description="""
    SNOMED CT is a clinical terminology with global scope covering a wide range of clinical specialties, disciplines and requirements. As a result of its broad scope, one of the benefits of SNOMED CT is a reduction of specialty boundary effects that arise from use of different terminologies or coding systems by different clinicians or departments. 
    This allows wider sharing and reuse of structured clinical information. Another benefit of SNOMED CT is that the same data can be processed and presented in ways that serve different purposes. For example, clinical records 
    represented using SNOMED CT can be processed and presented in different ways to support direct patient care, clinical audit, research, epidemiology, management and service planning. Additionally, the global scope of SNOMED CT reduces geographical boundary effects arising from the use of different terminologies or coding systems in different organizations and countries. [1]
    """,
    epilog="""
    [1] Snomed-CT. Retrieved May 9, 2019 from https://confluence.ihtsdotools.org/display/DOCSTART/3.+Using+SNOMED+CT+in+Clinical+Information
    """
)

parser_snomedct.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_snomedct.add_argument(
    "--versioned",
    dest='versioned',
    action='store_true',
    help="includes 'version' mapping"
)

parser_snomedct.add_argument(
    "--include-inactive",
    dest='include_inactive',
    action='store_true',
    help="includes 'inactive' relationship"
)

parser_snomedct.set_defaults(execute=SnomedCtCommand.execute)


parser_loinc = subparsers.add_parser(
    "loinc",
    help="Convert the 'LOINC' code system",
    description="""
    LOINC (Logical Observation Identifiers Names and Codes) was developed to provide a definitive standard for identifying clinical information in electronic reports. The LOINC database provides a set of universal names and ID codes for identifying laboratory and clinical test results in the context of existing HL7, ASTM E1238, and CEN TC251 observation report messages. One of the main goals of LOINC is to facilitate the exchange and pooling of results for clinical care, outcomes management, and research. LOINC codes are intended to identify the test result or clinical observation. Other fields in the message can transmit the identity of the source laboratory and special details about the sample. [1]
    """,
    epilog="""
    [1] "Knowledge Base - LOINC.“ Retrieved Sep 1, 2020 from https://loinc.org/kb/faq/basics/
    """
)

parser_loinc.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_loinc.add_argument(
    "--versioned",
    dest='versioned',
    action='store_true',
    help="include 'version'"
)

parser_loinc.set_defaults(execute=LoincCommand.execute)


parser_codesystem = subparsers.add_parser(
    "code-system",
    help="Convert a FHIR CodeSystem Resource",
    description="""
    The CodeSystem resource is used to declare the existence of and describe a code system or code system supplement and its key properties, and optionally define a part or all of its content.
    """,
    epilog="""
    [1] “4.8 Resource CodeSystem - Content.” CodeSystem - FHIR v4.0.0, Retrieved June 5, 2019 from www.hl7.org/fhir/codesystem.html.
    """
)

parser_codesystem.add_argument(
    "uri",
    metavar="URI",
    help="resource identifier for files"
)

parser_codesystem.set_defaults(execute=codesystem.execute)

args = parser.parse_args()

if hasattr(args, 'execute'):
    args.execute(args)
else:
    parser.print_help()




"""Verifies the 'configuration.py' module."""
import os

from uuid import uuid4 as uuid

from nose.tools import eq_, ok_

from termlink.configuration import Config


def test_configuration_is_loaded():
    """Checks that the configuration is loaded"""
    configuration = Config()
    ok_(configuration)


def test_logger_is_configured():
    """Checks that the application logger has been configured"""
    configuration = Config()
    ok_(configuration.logger)


def test_logger_level():
    """The level of the logger should be INFO for the DEFAULT environment"""
    configuration = Config(environment='DEFAULT')
    logger = configuration.logger
    eq_(20, logger.level)


def test_logger_level_for_test_environment():
    """The level of the logger should be DEBUG for the TEST environment"""
    configuration = Config(environment='TEST')
    logger = configuration.logger
    eq_(10, logger.level)


def test_get_os_property():
    """OS environment variables should be exposed in the config"""
    key = str(uuid())
    value = str(uuid())
    os.environ[key] = value
    configuration = Config()
    eq_(value, configuration.get_property(key))



"""Verifies the 'codesystem.py' module"""

from argparse import Namespace

import pkg_resources

from nose.tools import eq_, ok_, raises

from pronto import Term

from termlink.codesystem import _get_relationships, execute
from termlink.models import Coding, Relationship

@raises(ValueError)
def test_uri_scheme():
    """An unsupported URI scheme throws a ValueError"""
    execute(Namespace(uri='foo://bar'))

def test_json_format():
    """Tests the conversion of an .json file"""
    path = pkg_resources.resource_filename(__name__, "resources/codesystem.json")
    uri = f"file://{path}"
    output = execute(Namespace(uri=uri))
    ok_(len(output) > 0)



"""Handles Gene Sets conversion.

This module provides methods to extract, transform and load relationships
defined by the Geneset dataset.

The download files for Geneset are provided at http://software.broadinstitute.org/gsea/msigdb/collections.jsp.
"""

import csv
import json
from os import path

from re import match
from urllib.parse import urlparse

from termlink.models import Coding, Relationship, RelationshipSchema

_filename_regex = r'msigdb\..*\.symbols\.gmt'

def _to_relationship(rec, index, equivalence='subsumes'):
    """
    Convert record in table to Relationship as a JSON object

    Record is expected to have the following fields: [ source.CODE, source.STR,
    target.CODE, target.STR]

    Args:
        rec: A table record

    Returns:
        A new record containing a single field, which is the JSON object
    """
    source = Coding(
        system="http://www.broadinstitute.org/gsea/msigdb",
        code=rec[index],
        display=rec[index]
    )

    target = Coding(
        system="http://www.broadinstitute.org/gsea/msigdb",
        code=rec[0],
        display=rec[0]
    )

    return Relationship(equivalence, source, target)


def _get_relationships(uri):
    '''Extracts the system entities from the GSEA file

    Args:
        uri: a URI for the GSEA file on the local filesystem

    Returns:
        yields relationships
    '''
    with open(uri.path) as f:
        reader = csv.reader(f, delimiter='\t')
        for row in reader:
            for i in range(2, len(row)):
                yield _to_relationship(row, i)


def execute(args):
    '''Converts the GSEA ontology.

    Args:
        args:   command line arguments from argparse
    '''

    uri = urlparse(args.uri)
    if uri.scheme != 'file':
        raise ValueError(f"uri.scheme '{uri.scheme}' is not supported")

    filename = path.basename(uri.path)
    if not match(_filename_regex, filename):
        raise ValueError(
            f"File type is incorrect. Expected to match regular expression: '{_filename_regex}'. Found '{filename}'.")

    schema = RelationshipSchema()
    relationships = _get_relationships(uri)
    serialized = [ json.dumps(schema.dump(r)) for r in relationships ]

    for o in serialized:
        print(o)
        if "output" in args:
            with open(args.output, 'a') as f:
                f.write(o + '\n')
            
    return serialized



"""Handles LOINC conversion.

This module provides methods to extract, transform and load relationships
defined by the LOINC dataset.

The download files for LOINC are provided at https://loinc.org/downloads/.
"""
import glob
import os
import json
import re

from urllib.parse import urlparse

import petl as etl

from termlink.commands import SubCommand
from termlink.models import Coding, Relationship, RelationshipSchema

_SYSTEM = 'http://loinc.org'

def _find_loinc_table_core_file(path: str):
    path = os.path.join(path, '**/LoincTableCore.csv')
    paths = glob.glob(path)
    if not paths:
        raise RuntimeError("Unable to find LoincTableCore.csv")
    return paths[0]

def _find_multi_axial_hierarchy_file(path: str):
    path = os.path.join(path, '**/MultiAxialHierarchy.csv')
    paths = glob.glob(path)
    if not paths:
        raise RuntimeError("Unable to find MultiAxialHierarchy.csv")
    return paths[0]

def _parse_version(path: str):
    try:
        pattern = re.compile(r'Loinc_([\d.]*)_MultiAxialHierarchy')
        match = pattern.search(path)
        return match.group(1)
    except Exception as e:
        raise RuntimeError('Unable to parse version from path', path) from e

def _to_json(rec: etl.Record):
    """Converts a record to a formatted `Relationship` in JSON form.

    Args:
        rec: A table record

    Returns:
        A new record containing a single field, which is the JSON object
    """
    coding = Coding(
        system=_SYSTEM, 
        code=rec['source.CODE'],
        display=rec['source.CODE_TEXT'],
        version=rec.get('version')
    )

    parent = Coding(
        system=_SYSTEM, 
        code=rec['source.IMMEDIATE_PARENT'], 
        display=rec['target.CODE_TEXT'],
        version=rec.get('version')
    )

    relationships = [
        Relationship('subsumes', coding, parent),
        Relationship('specializes', parent, coding)
    ]

    schema = RelationshipSchema()
    return map(lambda relationship: [json.dumps(schema.dump(relationship))], relationships)


class Command(SubCommand):
    "A command executor for LOINC operations"

    @staticmethod
    def execute(args):
        """Prints a JSON array of `Relationship` objects to stdout

        Args:
            args: `argparse` parsed arguments
        """
        uri = urlparse(args.uri)
        service = Service(uri, args.versioned)
        table = service.get_relationships()
        etl.totext(table, source=None, encoding='utf8', template='{relationship}\n')


class Service:
    """Converts the LOINC Table Core files"""

    def __init__(self, uri, versioned):
        """Bootstraps a service

        Args:
            uri: URI to root location of the LOINC Table Core download files
        """

        if uri.scheme != 'file':
            raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

        self.uri = uri
        self.versioned = versioned

    def get_relationships(self):
        "Parses a list of `Relationship` objects."
        core_file = _find_loinc_table_core_file(self.uri.path)
        core = etl.fromcsv(core_file, delimiter=',')
        core = etl.cut(core, ['LOINC_NUM', 'LONG_COMMON_NAME'])
        hierarchy_file = _find_multi_axial_hierarchy_file(self.uri.path)
        hierarchy = etl.fromcsv(hierarchy_file, delimiter=',')
        hierarchy = etl.leftjoin(hierarchy, core, lkey='CODE', rkey='LOINC_NUM')
        hierarchy = etl.cut(hierarchy, ['IMMEDIATE_PARENT', 'CODE', 'CODE_TEXT', 'LONG_COMMON_NAME'])
        hierarchy = etl.fillright(hierarchy)
        hierarchy = etl.cut(hierarchy, ['IMMEDIATE_PARENT', 'CODE', 'LONG_COMMON_NAME'])
        hierarchy = etl.rename(hierarchy, 'LONG_COMMON_NAME', 'CODE_TEXT')
        parents = etl.cut(hierarchy, ['CODE', 'CODE_TEXT'])
        hierarchy = etl.selectne(hierarchy, 'IMMEDIATE_PARENT', '')
        hierarchy = etl.leftjoin(hierarchy, parents, lkey='IMMEDIATE_PARENT', rkey='CODE', lprefix='source.', rprefix='target.')
        hierarchy = etl.distinct(hierarchy)
        if self.versioned:
            version = _parse_version(hierarchy_file)
            hierarchy = etl.addfield(hierarchy, 'version', version)
        hierarchy = etl.rowmapmany(hierarchy, _to_json, ['relationship'])
        return hierarchy



"""
Utility method for loading properties form the configuration file and environment variables.
"""

import configparser
import logging
import os

# Relative location of configuration file
_PROGRAM_CONFIG = '../config.ini'
_LOGGER_FILENAME = 'stderr.log'


class Config:
    """A configuration supplier.

    An object for reading configuration properties. Properties are injected with
    the following priority. Properties with higher priority overwrite those
    with lower priority.

    1. Program defaults
    2. The 'DEFAULT' section in 'config.ini'
    3. The environment in the 'config.ini' file set by the 'ENV' environment variable
    4. Runtime system environment variables
    """

    def __init__(self, environment=os.getenv("ENV", 'DEFAULT').upper()):

        # Create an application logger
        logger = logging.getLogger("termlink")
        logging.basicConfig(filename=_LOGGER_FILENAME)

        # Set the logging level
        if environment == "TEST":
            logger.setLevel(logging.DEBUG)
        else:
            logger.setLevel(logging.INFO)

        logger.info("The logging level has been set to %s", logger.level)

        parser = configparser.ConfigParser(os.environ, strict=False)

        # Get the absolute path based on the execution directory
        root = os.path.abspath(os.path.dirname(__file__))
        config = os.path.join(root, _PROGRAM_CONFIG)

        configs = [config]
        parser.read(configs)

        self.environment = environment
        self.logger = logger
        self.parser = parser

    def get_property(self, property_name, default=None):
        """Get a property value from the configuration.

        The os environment variables are first checked. If the property
        does not exists there it is read from the configuration file.

        Args:
            property_name (str):    A name, or key, of a property

        Returns:
            The property value associated with the property_name
        """
        if property_name in os.environ:
            return os.environ[property_name]

        return self.parser.get(self.environment, property_name, fallback=default)



"""Verifies the 'common.py' module"""

from argparse import Namespace

import pkg_resources

from nose.tools import eq_, ok_, raises

from pronto import Term

from termlink.common import _to_coding, _to_equivalence_from_scope, _to_relationship, execute
from termlink.models import Coding, Relationship


@raises(ValueError)
def test_uri_scheme():
    """An unsupported URI scheme throws a ValueError"""
    execute(Namespace(uri='foo://bar'))


def test_obo_format():
    """Tests the conversion of an .obo file"""
    path = pkg_resources.resource_filename(__name__, "resources/ontology.obo")
    uri = f"file://{path}"
    system = 'https://lifeomic.github.io/termlink/'
    output = execute(Namespace(uri=uri, system=system))
    ok_(len(output) == 4)


def test_owl_format():
    """Tests the conversion of an .owl file"""
    path = pkg_resources.resource_filename(__name__, "resources/ontology.owl")
    uri = f"file://{path}"
    system = 'https://lifeomic.github.io/termlink/'
    output = execute(Namespace(uri=uri, system=system))
    ok_(len(output) == 2)


def test_to_coding():
    """Checks that a term is properly converted"""

    system = "http://snomed.info/sct"
    term = Term(id='SNOMEDCT_US:25064002', name='Headache')

    res = _to_coding(term, system)

    exp = Coding(
        system=system,
        code='25064002',
        display='Headache'
    )

    eq_(exp, res)


def test_to_coding_without_colon():
    """Checks that a term without a ':' is properly converted"""

    system = "http://snomed.info/sct"
    term = Term(id='25064002', name='Headache')

    res = _to_coding(term, system)

    exp = Coding(
        system=system,
        code='25064002',
        display='Headache'
    )

    eq_(exp, res)

def test_to_coding_with_system_and_code():
    """Checks that a term with a system and code identifier is properly converted"""

    system = "http://snomed.info/sct"
    term = Term(id='http://snomed.info/sct:25064002', name='Headache')

    res = _to_coding(term, system)

    exp = Coding(
        system=system,
        code='25064002',
        display='Headache'
    )

    eq_(exp, res)

def test_to_coding_with_system():
    """Checks that a term with a system identifier is properly converted"""

    system = "http://snomed.info/sct"
    term = Term(id='http://snomed.info/sct:', name='SNOMED-CT')

    res = _to_coding(term, system)

    exp = Coding(
        system=system,
        display='SNOMED-CT'
    )

    eq_(exp, res)


def test_to_json():
    """Checks that a source, equivalence and target and properly converted"""

    system = "http://snomed.info/sct"
    source = Term(id='SNOMEDCT_US:735938006', name='Acute headache')
    equivalence = 'subsumes'
    target = Term(id='SNOMEDCT_US:25064002', name='Headache')

    res = _to_relationship(source, equivalence, target, system)

    exp = Relationship(
        equivalence='subsumes',
        source=Coding(
            system=system,
            code='735938006',
            display='Acute headache'
        ),
        target=Coding(
            system=system,
            code='25064002',
            display='Headache'
        )
    )

    eq_(exp, res)

@raises(RuntimeError)
def test_to_equivalence_from_scope():
    """A RuntimeError is thrown for an unknown scope"""
    _to_equivalence_from_scope("foobar")




"""Verifies the 'gsea.py' module"""

from argparse import Namespace

import pkg_resources
import tempfile

from nose.tools import eq_, ok_, raises

from termlink.gsea import execute, _to_relationship
from termlink.models import Coding, Relationship

def test_execute():
    "Tests the conversion of a 'msigdb.*.symbols.gmt' file"
    path = pkg_resources.resource_filename(__name__, "resources/msigdb.test.symbols.gmt")
    uri = f"file://{path}"
    output = execute(Namespace(uri=uri))
    ok_(len(output) > 0)

def test_execute_and_write_to_file():
    "Tests the conversion of a 'msigdb.*.symbols.gmt' file"
    path = pkg_resources.resource_filename(__name__, "resources/msigdb.test.symbols.gmt")
    uri = f"file://{path}"
    (_, tmp) = tempfile.mkstemp()
    output = execute(Namespace(uri=uri, output=tmp))
    ok_(len(output) > 0)

@raises(ValueError)
def test_uri_scheme():
    """A URI scheme of 'file://' is required"""
    execute(Namespace(uri='foo://bar'))


@raises(ValueError)
def test_uri_filename_database():
    "A invalid file name throws a ValueError"
    execute(Namespace(uri='file:///invalid.foo.bar.symbols.gmt'))


@raises(ValueError)
def test_uri_filename_type():
    "A invalid file name throws a ValueError"
    execute(Namespace(uri='file:///msigdb.foo.bar.invalid.gmt'))


@raises(ValueError)
def test_uri_filename_suffix():
    "A invalid file name throws a ValueError"
    execute(Namespace(uri='file:///msigdb.foo.bar.symbols.invalid'))


def test_to_relationship():
    "A record is converted to an index"
    rec = ['target', 'source']
    idx = 1
    relationship = _to_relationship(rec, idx, equivalence='subsumes')
    eq_(relationship, Relationship(
        equivalence='subsumes',
        source=Coding(
            system="http://www.broadinstitute.org/gsea/msigdb",
            code='source',
            display='source'
        ),
        target=Coding(
            system="http://www.broadinstitute.org/gsea/msigdb",
            code='target',
            display='target'
        )
    ))




"""Verifies the 'rxnorm.py' module"""

from urllib.parse import urlparse

from nose.tools import ok_, raises, eq_

from petl import Record

from termlink.models import Coding, Relationship
from termlink.rxnorm import Service, _to_json, _to_system, _to_relationship


def test_service_uri_can_be_file():
    """Checks that a uri.scheme of 'file' is ok"""
    uri = urlparse("file://")
    ok_(Service(uri=uri))


@raises(ValueError)
def test_service_uri_requires_scheme_file():
    """Checks that a uri.scheme of 'foobar' throws a ValueError"""
    uri = urlparse("foobar://")
    Service(uri=uri)


def test_to_relationship():
    """Checks that a record is properly converted to a Relationship"""
    fields = ['REL', 'source.CODE', 'source.STR', 'source.SAB',
              'target.CODE', 'target.STR', 'target.SAB']
    values = (
        'RB',
        '313782',
        'Acetaminophen 325 MG Oral Tablet',
        'RXNORM',
        '1152843',
        'Acetaminophen Pill',
        'RXNORM'
    )

    rec = Record(values, fields)
    res = _to_relationship(rec)
    exp = Relationship(
        equivalence='subsumes',
        source=Coding(
            system='http://www.nlm.nih.gov/research/umls/rxnorm',
            code='313782',
            display='Acetaminophen 325 MG Oral Tablet'
        ),
        target=Coding(
            system='http://www.nlm.nih.gov/research/umls/rxnorm',
            code='1152843',
            display='Acetaminophen Pill'
        )
    )

    eq_(exp, res)


def test_to_json():
    """Checks that a record is properly converted to .json"""

    fields = ['REL', 'source.CODE', 'source.STR', 'source.SAB',
              'target.CODE', 'target.STR', 'target.SAB']
    values = (
        'RB',
        '313782',
        'Acetaminophen 325 MG Oral Tablet',
        'RXNORM',
        '1152843',
        'Acetaminophen Pill',
        'RXNORM'
    )

    src = Record(values, fields)
    res = _to_json(src)
    ok_(res)


def test_to_system():
    """Checks that an sab is correctly converted into a system"""
    sab = 'rxnorm'
    res = _to_system(sab)
    exp = 'http://www.nlm.nih.gov/research/umls/rxnorm'
    eq_(exp, res)



"""Handles FHIR CodeSystem Resource conversion.


This modules provides methods to extract, transform and load relationship
defined by a FHIR CodeSystem resource.

Various FHIR CodeSystems are available at https://www.hl7.org/fhir/codesystem.html
"""

import json

from urllib.parse import urlparse

from termlink.models import Coding, Relationship, RelationshipSchema

def _get_relationships(content):
    '''Extracts concept relationships from a CodeSystem

    Args:
        content: a dict of a CodeSystem

    Returns:
        yields relationships
    '''
    
    system = content['url']
    version = content['version']

    sources = content['concept']
    targets = [sources[-1]] + sources[0:-1]
    for source, target in zip(sources, targets):
        
        source = Coding(
            system=system,
            version=version,
            code=source['code'],
            display=source['display']
        )

        target = Coding(
            system=system,
            version=version,
            code=target['code'],
            display=target['display']
        )

        yield Relationship('disjoint', source, target)

def execute(args):
    '''Converts a FHIR CodeSystem Resource

    Args:
        args:   command line arguments from argparse
    '''
    uri = urlparse(args.uri)
    if uri.scheme != 'file':
        raise ValueError(f"uri.scheme '{uri.scheme}' is not supported")

    with open(uri.path) as f:
        content = json.load(f)

    schema = RelationshipSchema()
    relationships = _get_relationships(content)
    serialized = [json.dumps(schema.dump(r)) for r in relationships]

    for o in serialized:
        print(o)

    return serialized

    



"""Handles Human Phenotype Ontology conversion.

This module provides methods to extract and transform the HPO (Human Phenotype Ontology).

The download files for HPO are provided at https://hpo.jax.org/app/download/ontology
"""
import json

from urllib.parse import urlparse

from pronto import Ontology, Term

from termlink.commands import SubCommand
from termlink.models import Coding, Relationship, RelationshipSchema
from termlink.services import RelationshipService


def _to_equivalence_from_scope(scope):
    """Converts a scope into an equivalence

    Args:
        scope: a `pronto.Synonym.scope`

    Returns:
        an equivalence from https://www.hl7.org/fhir/valueset-concept-map-equivalence.html
    """
    try:
        return {
            'BROAD': 'wider',
            'NARROW': 'narrower',
            'EXACT': 'equivalent',
            'RELATED': 'relatedto'
        }[scope]
    except KeyError:
        raise RuntimeError('scope \'%s\' is not supported' % scope)


def _to_system(abbreviation):
    """Converts an abbreviation to a system identifier.

    Args:
        abbreviation: a `pronto.Term.id`

    Returns:
        a system identifier
    """
    try:
        return {
            'HP': 'http://www.human-phenotype-ontology.org/'
        }[abbreviation]
    except KeyError:
        raise RuntimeError(
            'system abbreviation \'%s\' is not supported' % abbreviation)


def _to_coding(term):
    """Converts a term into a `Coding`.

    Args:
        term: A `pronto.Term`

    Returns:
        a `termlink.models.Coding`
    """
    (abbreviation, code) = term.id.split(':')
    return Coding(
        system=_to_system(abbreviation),
        code=code,
        display=term.name
    )


def _to_relationship(source, equivalence, target):
    """Converts a source and target `pronto.Term` into a JSON object.

    Args:
        source: a `pronto.Term`
        equivalence: a concept map equivalence
        target: a `pronto.Term`

    Returns:
        a 'termlink.models.Relationship` in JSON form
    """
    source = _to_coding(source)
    target = _to_coding(target)
    return Relationship(equivalence, source, target)


class Command(SubCommand):
    "A command executor for Human Phenotype Ontology operations."
    @staticmethod
    def execute(args):
        uri = urlparse(args.uri)
        skip_synonyms = args.skip_synonyms
        skip_alt_ids = args.skip_alt_ids

        service = Service(uri, skip_alt_ids, skip_synonyms)

        relationships = service.get_relationships()
        schema = RelationshipSchema()
        relationships = [schema.dump(relationship)
                         for relationship in relationships]
        print(json.dumps(relationships))


class Service(RelationshipService):
    "Converts the Human Phenotype Ontology"

    def __init__(self, uri, skip_alt_ids=False, skip_synonyms=False):
        """Bootstraps a service

        Args:
            uri: URI to root location of .obo files
            skip_alt_ids: Skips 'alt_id' conversions if True
            skip_synonyms: Skips 'synonym' conversions if True
        """
        if uri.scheme != 'file':
            raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

        self.uri = uri
        self.skip_synonyms = skip_synonyms
        self.skip_alt_ids = skip_alt_ids

    def get_relationships(self):
        """Parses a list of `Relationship` objects

        Returns:
            yields `Relationship`s in JSON form
        """
        ontology = Ontology(self.uri.path)

        # child to parent relationships
        for term in ontology:
            for child in term.children:
                yield _to_relationship(child, "subsumes", term)

        # parent to child relationships
        for term in ontology:
            for parent in term.parents:
                yield _to_relationship(parent, "specializes", term)

        # alt_id relationships:
        if not self.skip_alt_ids:
            for term in ontology:
                for other, values in term.other.items():
                    if other == 'alt_id':
                        for value in values:
                            target = Term(
                                id=value,
                                name=term.name,
                                desc=term.desc
                            )
                            yield _to_relationship(term, "equal", target)

        # synonym relationships
        if not self.skip_synonyms:
            for term in ontology:
                for synonym in term.synonyms:
                    target = Term(
                        id=term.id,
                        name=synonym.desc,
                        desc=term.desc
                    )
                    equivalence = _to_equivalence_from_scope(synonym.scope)
                    yield _to_relationship(term, equivalence, target)

        # todo - lookup concept based on system and code
        # xref relationships
        # for term in ontology:
        #     for other, values in term.other.items():
        #         if other == 'xref':
        #             for value in values:
        #                 (abbreviation, code) = value.split(':')
        #                 system = _to_system(abbreviation)



"""Model representations of API interfaces"""

from dataclasses import dataclass

from marshmallow import Schema, fields

from termlink.configuration import Config

_configuration = Config()
_logger = _configuration.logger


class CodingSchema(Schema):
    """A `Coding` schema

    Attributes:
        type:       string field
        system:     string field
        version:    string field
        code:       string field
        display     string field
    """
    type = fields.Str()
    system = fields.Str()
    version = fields.Str()
    code = fields.Str()
    display = fields.Str()


@dataclass(frozen=True)
class Coding():
    """
    A 'Coding' object as defined by the API.

    Attributes:
        type (str):     the type of object
        system (str):   identity of the terminology system
        version (str):  version of the system
        code (str):     symbol in syntax defined by the system
        display (str):  representation defined by the system
    """
    type: str = 'coding'
    system: str = None
    version: str = None
    code: str = None
    display: str = None


class RelationshipSchema(Schema):
    """A `Relationship` schema

    Attributes:
        system: string field
        source: `CodingSchema` field
        target: `CodingSchema` field
    """
    equivalence = fields.Str()
    source = fields.Nested(CodingSchema())
    target = fields.Nested(CodingSchema())


@dataclass(frozen=True)
class Relationship():
    """
    A 'Relationship' object as defined by the API.

    Attributes:
        equivalence (str):  The degree of equivalence between concepts.
        source object:       The source concept.
        target object:       The target concept.
    """
    equivalence: str
    source: Coding
    target: Coding



"""Handles conversions of common ontology formats

This module provides support for converting common ontology formats. The
following file types are supported:

- .obo: https://owlcollab.github.io/oboformat/doc/GO.format.obo-1_4.html
- .owl: https://www.w3.org/OWL/

"""

import json

from urllib.parse import urlparse

import validators

from pronto import Ontology

from termlink.models import Coding, Relationship, RelationshipSchema

_SCOPE_TO_INVERSE_SCOPE = {
    'equivalent_to': 'equivalent_to'
}

_SCOPE_TO_EQUIVALENCE = {
    'equivalent_to': 'equivalent'
}

def _to_inverse_scope(scope):
    """Converts a scope into its inverse scope

    Args:
        scope: a `pronto.Synonym.scope`

    Returns:
        an inverse scope
    """
    try:
        return _SCOPE_TO_INVERSE_SCOPE[scope]
    except KeyError:
        raise RuntimeError('scope \'%s\' is not supported' % scope)

def _to_equivalence_from_scope(scope):
    """Converts a scope into an equivalence

    Args:
        scope: a `pronto.Synonym.scope`

    Returns:
        an equivalence from https://www.hl7.org/fhir/valueset-concept-map-equivalence.html
    """
    try:
        return _SCOPE_TO_EQUIVALENCE[scope]
    except KeyError:
        raise RuntimeError('scope \'%s\' is not supported' % scope)


def _to_coding(term, system):
    """Converts a term into a `Coding`.

    Args:
        term: A `pronto.Term`

    Returns:
        a `termlink.models.Coding`
    """
    _id = term.id
    if ':' in _id:
        if _id.endswith(':') and validators.url(_id[:-1]):
            system = _id[:-1]
            code = None
        else:
            parts = _id.rsplit(':', 1)
            if validators.url(parts[0]):
                system = parts[0]
                code = parts[1]
            else:
                system = system
                code = parts[1]
    else:
        system = system
        code = _id

    return Coding(
        system=system,
        code=code,
        display=term.name
    )


def _to_relationship(source, equivalence, target, system):
    """Converts a source and target `pronto.Term` into a JSON object.

    Args:
        source: a `pronto.Term`
        equivalence: a concept map equivalence
        target: a `pronto.Term`

    Returns:
        a 'termlink.models.Relationship` in JSON form
    """
    source = _to_coding(source, system)
    target = _to_coding(target, system)
    return Relationship(equivalence, source, target)


def _get_relationships(uri, system):
    """Parses a list of `Relationship` objects

    Args:
        uri:    a URI for the ontology file on the local filesystem
        system: the target system

    Returns:
        yields relationships
    """
    ontology = Ontology(uri.path)

    # child to parent relationships
    for term in ontology:
        for child in term.children:
            yield _to_relationship(child, "subsumes", term, system)

    # parent to child relationships
    for term in ontology:
        for parent in term.parents:
            yield _to_relationship(parent, "specializes", term, system)

    for term in ontology:
        for scope, references in term.other.items():
            if scope in _SCOPE_TO_EQUIVALENCE:
                for reference in references:
                    relationship = _to_equivalence_from_scope(scope)
                    yield _to_relationship(term, relationship, ontology[reference], system)
            if scope in _SCOPE_TO_INVERSE_SCOPE:
                inverse = _to_inverse_scope(scope)
                for reference in references:
                    relationship = _to_equivalence_from_scope(inverse)
                    yield _to_relationship(ontology[reference], relationship, term, system)


def execute(args):
    """
    Converts an ontology in a common format.

    Args:
        args:   command line arguments from argparse
    """
    uri = urlparse(args.uri)
    if uri.scheme != 'file':
        raise ValueError("'uri.scheme' %s not supported" % uri.scheme)

    schema = RelationshipSchema()
    relationships = _get_relationships(uri, args.system)
    serialized = [json.dumps(schema.dump(r)) for r in relationships]

    for o in serialized:
        print(o)

    return serialized



"""Verifies the 'hpo.py' module"""

from urllib.parse import urlparse

from nose.tools import ok_, raises, eq_

from parameterized import parameterized

from pronto import Term

from termlink.hpo import Service, _to_equivalence_from_scope, _to_system, _to_coding, _to_relationship
from termlink.models import Coding, Relationship


def test_service_uri_can_be_file():
    """Checks that a uri.scheme of 'file' is ok"""
    uri = urlparse("file://")
    ok_(Service(uri=uri))


@raises(ValueError)
def test_service_uri_requires_scheme_file():
    """Checks that a uri.scheme of 'foobar' throws a ValueError"""
    uri = urlparse("foobar://")
    Service(uri=uri)


@parameterized([
    ('BROAD', 'wider'),
    ('NARROW', 'narrower'),
    ('EXACT', 'equivalent'),
    ('RELATED', 'relatedto')
])
def test_to_equivalence_from_scope(scope, exp):
    """Checks that each scope converts"""
    res = _to_equivalence_from_scope(scope)
    ok_(exp == res)


@raises(RuntimeError)
def test_to_equivalence_from_scope_throws_runtime_error():
    """Checks that a runtime error is thrown for invalid scope"""
    _to_equivalence_from_scope('foobar')


@parameterized([
    ('HP', 'http://www.human-phenotype-ontology.org/')
])
def test_to_system(abbreviation, exp):
    """Checks that each scope converts"""
    res = _to_system(abbreviation)
    ok_(exp == res)


@raises(RuntimeError)
def test_to_system_throws_runtime_error():
    """Checks that a runtime error is thrown for invalid abbrevation"""
    _to_system('foobar')


def test_to_coding():
    """Checks that a term is properly converted to a coding"""

    term = Term(id='HP:0000001', name='All')

    res = _to_coding(term)

    exp = Coding(
        system='http://www.human-phenotype-ontology.org/',
        code='0000001',
        display='All'
    )

    ok_(exp == res)


def test_to_relationship():
    """Checks that a source, equivalence and target and properly converted"""

    source = Term(id='HP:0000107', name='Renal cyst')
    equivalence = 'subsumes'
    target = Term(id='HP:0000003', name='Multicystic kidney dysplasia')

    res = _to_relationship(source, equivalence, target)

    exp = Relationship(
        equivalence='subsumes',
        source=Coding(
            system='http://www.human-phenotype-ontology.org/',
            code='0000107',
            display='Renal cyst'
        ),
        target=Coding(
            system='http://www.human-phenotype-ontology.org/',
            code='0000003',
            display='Multicystic kidney dysplasia'
        )
    )

    eq_(exp, res)
